/************************************************************************
	zsuperlu_rcf.c  -- Row-Compressed-Format
	-- For Complex data only.

	Copied from ../EXAMPLE/superlu_rcf.c
	Modified by G. Shi, 10-23-02, for testing some circuit systems.
	Data (A,b) are provided by files in the format of 
	  (row, col, val) -- Real
	  (row, col, r_val, i_val) -- Complex
*************************************************************************/

/*
 * -- SuperLU routine (version 2.0) --
 * Univ. of California Berkeley, Xerox Palo Alto Research Center,
 * and Lawrence Berkeley National Lab.
 * November 15, 1997
 *
 */

#include <stdio.h>

/*#include "dsp_defs.h"*/

#include "zsp_defs.h"
#include "util.h"

#include "ctype.h"

#define		No		0
#define		Yes 	1

static char *FileName;

int		m, n, nnz = 0;
double 	*a, *rhs;  			/* Real data */
doublecomplex *za, *zrhs;  	/* Complex data */
int     *colind, *rowptr;  	/* Used in Compressed Row Format (CRF). */

int		Complex = Yes;

static int  ReadMatrixFromFile();



main(int argc, char *argv[])
{
/*
 * Purpose
 * =======
 * 
 * This is the small 5x5 example used in the Sections 1 and 2 of the 
 * User's Guide to illustrate how to call a SuperLU routine, and the
 * matrix data structures used by SuperLU.
 *
 */

    SuperMatrix A, L, U, B;
    double   s, u, p, e, r, l;
    int      *perm_r; /* row permutations from partial pivoting */
    int      *perm_c; /* column permutation vector */
    int      nrhs = 1, info, i, permc_spec;

	/* Newly added variables */
	int		Error;
	double  *at;		 	/* Stores the transpose of 'a' */
	doublecomplex  *zat;	/* Stores the transpose of 'za' */
	int     *rowind, *colptr;  /* Used in Compressed Column Format (CCF)


	/* Read matrix file. */
	if (argc == 1) {
		fprintf(stderr, "Error: No data file.\nUsage: superlu data_file.\n");
		exit(1);
	}
	else 
		FileName = *(++argv);

	Error = ReadMatrixFromFile();

	if (Error) {
		fprintf(stderr, "Error: Failed reading data file. \n ");
		exit(1);
	}

	/* Allocate memory for CCF. */

	/* Not needed !!
    if ( !(at = doubleMalloc(nnz)) ) ABORT("Malloc fails for at[].");
    if ( !(rowind = intMalloc(nnz)) ) ABORT("Malloc fails for rowind[].");
    if ( !(colptr = intMalloc(n+1)) ) ABORT("Malloc fails for colptr[].");
	*/

	/* Convert from Compressed Row Format to Compressed Column Format. */
	zCompRow_to_CompCol(m, n, nnz, za, colind, rowptr, &zat, &rowind, &colptr);

	/* This function is defined in '../SRC/dutil.c and dutil.h'. */

	/* Free memory used by CRF: 'a, colind, rowptr' */
	free(za);
	free(colind);
	free(rowptr);

	/* Print to test */
	/*	
	printf("\nzat: \n");
	for (i = 0; i < nnz; i++)   printf(" %.2lf + j %.2lf;", zat[i].r, zat[i].i);
	printf("\nrowind : \n");
	for (i = 0; i < nnz; i++)   printf(" %d ", rowind[i]);
	printf("\ncolptr: \n");
	for (i = 0; i < n+1; i++)   printf(" %d ", colptr[i]);
	printf("\nzrhs: \n");
	for (i = 0; i < n; i++)   printf(" %.2lf + j %.2lf;", zrhs[i].r, zrhs[i].i);
	*/	
	printf("\n");

	/*printf("\n"); exit(1);*/ 
	/* Terminate the program here. */


    /* Create matrix A in the format expected by SuperLU. */
	/* Note: Must use 'zat' */
    zCreate_CompCol_Matrix(&A, m, n, nnz, zat, rowind, colptr, NC, Z_, GE);
	/* 'Z_' = double-complex */
	fprintf(stderr, "Matrix created.\n"); fflush(stderr);
    
    /* Create right-hand side matrix B. */
    zCreate_Dense_Matrix(&B, m, nrhs, zrhs, m, DN, Z_, GE);
	fprintf(stderr, "RHS created.\n");  fflush(stderr);

    if ( !(perm_r = intMalloc(m)) ) ABORT("Malloc fails for perm_r[].");
    if ( !(perm_c = intMalloc(n)) ) ABORT("Malloc fails for perm_c[].");

    /*
     * Get column permutation vector perm_c[], according to permc_spec:
     *   permc_spec = 0: use the natural ordering 
     *   permc_spec = 1: use minimum degree ordering on structure of A'*A
     *   permc_spec = 2: use minimum degree ordering on structure of A'+A
     */    	
    permc_spec = 1;   /* Options: 0, 1, 2 */

    get_perm_c(permc_spec, &A, perm_c);

    zgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
	fprintf(stderr, "Finished solving.\n"); fflush(stderr);
    
	/* Check the 'info'. */
	printf("\ninfo = %d ", info);   /* info = 0 : successful exit */
	if (info == 0)  printf(": Successful exit. \n");
	else            printf(": Unsuccessful exit. \n");

	/* Print the solution which is in 'B'. */
    zPrint_Dense_Matrix("B", &B);
	
	/*
    dPrint_CompCol_Matrix("A", &A);
    dPrint_CompCol_Matrix("U", &U);
    dPrint_SuperNode_Matrix("L", &L);
    print_int_vec("\nperm_r", m, perm_r);
	*/	

    /* De-allocate storage */
    SUPERLU_FREE (zat);
    SUPERLU_FREE (zrhs);

    SUPERLU_FREE (rowind);
    SUPERLU_FREE (colptr);
	/* Above three lines newly added. */

    SUPERLU_FREE (perm_r);
    SUPERLU_FREE (perm_c);
    Destroy_CompCol_Matrix(&A);
    Destroy_SuperMatrix_Store(&B);
    Destroy_SuperNode_Matrix(&L);
    Destroy_CompCol_Matrix(&U);
}


static int  ReadMatrixFromFile()
{
	FILE  	*pMatrixFile;
	int   	Error, Row, Col, Count = 0, LineNumber, RHS_Col;
	int	  	Size = 0;
	long  	Reads;
	double 	Real, Imag;
	char  	sInput[BUFSIZ], sType[BUFSIZ];
	char	*p;
	static char Message[BUFSIZ];

	static char *EndOfFile = "Unexpected end of file '%s' at line %d.\n";
	static char *Syntax = "Syntax error in file '%s' at line %d.\n";
	static char *Index = "Index error in file '%s' at line %d.\n";

	int		i, j;    
	/* Indices used to trace the three arrays in Compressed Row Format. */
	int		curr_row;		/* The current row index. */

	pMatrixFile = fopen(FileName, "r");
	if (pMatrixFile == NULL) {
		fprintf(stderr, "Failed to open file: %s \n", FileName);
		return 1;
	}
	
	LineNumber = 1;

	/* Read and print label. */
	if (NULL == fgets(Message, BUFSIZ, pMatrixFile)) {
		fprintf(stderr, EndOfFile, FileName, LineNumber);
		return 1;
	}

	/* Check by print */
	/* printf(" Message = %s \n", Message); */

	/* Read the size of matrix. */
	LineNumber++;

	if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
		fprintf(stderr, EndOfFile, FileName, LineNumber);
		return 1;
	}
	if ((Reads = sscanf(sInput, "%d %s", &Size, sType)) < 1) {
		fprintf(stderr, Syntax, FileName, LineNumber);
		return 1;
	}
	if (Reads == 2) {
		for (p = sType; *p != '\n'; p++)
			*p = tolower(*p);
		if (strncmp( sType, "complex", 7) == 0)
			Complex = Yes;
		else if (strncmp( sType, "real", 4) == 0)
			Complex = No;
		else {
			fprintf(stderr, Syntax, FileName, LineNumber);
			return 1;
		}
	}

	/* Print to Check */
	/* printf(" Reads = %d \n", Reads); */

	/* Set the size of matrix A. (m, n) are globals. */
	m = n = Size;

	/* Read data file once to get the nonzero count (nnz). */
	nnz = 0;	/* Initialize num_of_nonzeros */
	do {
		LineNumber++;

		if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
			fprintf(stderr, EndOfFile, FileName, LineNumber);
			return 1;
		}

		if (Complex) {
			Reads = sscanf(sInput, "%d%d%lf%lf", &Row, &Col, &Real, &Imag);
			if (Reads != 4) {
				fprintf(stderr, Syntax, FileName, LineNumber);
				return 1;
			}
		} 
		else {  /* Real */
			Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);
			if (Reads != 3) {
				fprintf(stderr, Syntax, FileName, LineNumber);
				return 1;
			}
		}

		if (Row < 0 || Row > Size || Col < 0 || Col > Size) {
			fprintf(stderr, Index, FileName, LineNumber);
			return 1;
		}

		/* Print to check */
		/* printf(" %d, %d, %.2lf \n", Row, Col, Real); */
		/* printf(" %d, %d, %.2lf, %.2lf \n", Row, Col, Real, Imag); */

		/* Get the total num of nonzeros so that memory can be allocated. */
		if (Row > 0 && Col > 0)		++nnz;

	} while (Row != 0 && Col != 0);

	/* Print to check */
	printf("\nSize = %d,  Type = %s \n", Size, sType);
	printf("nnz = %d \n", nnz); 
	printf("sparsity (nnz/n) = %.6lf \n", (double) nnz/Size); 

	/* Allocate memory for Compressed-Row-Format and RHS. */

	if ( Complex ) {
    	if ( !(za = doublecomplexMalloc(nnz)) ) 
			ABORT("Malloc fails for za[].");
    	if ( !(zrhs = doublecomplexMalloc(m)) ) 
			ABORT("Malloc fails for zrhs[].");
	}
	else {
    	if ( !(a = doubleMalloc(nnz)) ) 
			ABORT("Malloc fails for a[].");
    	if ( !(rhs = doubleMalloc(m)) ) 
			ABORT("Malloc fails for rhs[].");
	}

    if ( !(colind = intMalloc(nnz)) ) ABORT("Malloc fails for colind[].");
    if ( !(rowptr = intMalloc(n+1)) ) ABORT("Malloc fails for rowptr[].");


    /* Read data file again to set up arrays in Compressed-Row-Format. */

	/* Resume the file pointer */
	rewind(pMatrixFile);

	/* Skip the first two lines. */
	fgets(Message, BUFSIZ, pMatrixFile);
	fgets(Message, BUFSIZ, pMatrixFile);
	LineNumber = 2;

	i = j = 0;
	curr_row = 0;		/* row index counts from 1 in the data file */

    while (1) {

		LineNumber++;
    	fgets(sInput, BUFSIZ, pMatrixFile);
		if (Complex ) {
    		Reads = sscanf(sInput, "%d%d%lf%lf", &Row, &Col, &Real, &Imag);
			if (Row == 0) 	break;

			za[i].r = Real;			/* real part */
			za[i].i = Imag;			/* imaginary part */

		}
		else {
    		Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);
			if (Row == 0) 	break;

			a[i] = Real;			/* nonzero value */
		}
		colind[i] = Col - 1;	/* colind = column index */

		if (Row > curr_row) {
			rowptr[j++] = i; 
			curr_row++;
		}
		++i;
    } 

	rowptr[j] = i;		/* Append the ending index in rowptr. */

	/* Print to test */
	/*	
	printf("\nza: \n");
	for (i = 0; i < nnz; i++)   printf(" %.2lf + j %.2lf;", za[i].r, za[i].i);
	printf("\ncolind : \n");
	for (i = 0; i < nnz; i++)   printf(" %d ", colind[i]);
	printf("\nrowptr: \n");
	for (i = 0; i < n+1; i++)   printf(" %d ", rowptr[i]);
	*/	

	/* Continue to set up the RHS */
	for (i = 0; i < Size; i++) {
		LineNumber++;

		if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
			fprintf(stderr, EndOfFile, FileName, LineNumber);
			return 1;
		}

		if (Complex) {
			Reads = sscanf(sInput, "%lf%lf", &Real, &Imag);
			if (Reads != 2) {
				fprintf(stderr, Syntax, FileName, LineNumber);
				return 1;
			}
			zrhs[i].r = Real;
			zrhs[i].i = Imag;
		}
		else {
			Reads = sscanf(sInput, "%lf", &Real);
			if (Reads != 1) {
				fprintf(stderr, Syntax, FileName, LineNumber);
				return 1;
			}
			rhs[i] = Real;
		}
	}

	/* Print to test */
	/*	
	printf("\nrhs: \n");
	for (i = 0; i < m; i++)   printf(" %.2lf + j %.2lf;", zrhs[i].r, zrhs[i].i);
	*/	


	(void) fclose(pMatrixFile);

	printf("\n");

	return 0;		/* No error. */
}

